package digi.recipeManager;

import java.util.HashMap;

import org.bukkit.*;
import org.bukkit.block.*;
import org.bukkit.entity.Player;
import org.bukkit.event.Event.Result;
import org.bukkit.event.*;
import org.bukkit.event.block.BlockBreakEvent;
import org.bukkit.event.inventory.*;
import org.bukkit.event.player.PlayerQuitEvent;
import org.bukkit.event.world.*;
import org.bukkit.inventory.*;
import org.bukkit.inventory.Recipe;

import digi.recipeManager.customEvents.*;
import digi.recipeManager.data.*;

public final class Events implements Listener
{
	private final RecipeManager	plugin;
	private final Recipes		recipes;
	private final Econ			economy;
	private final Settings		settings;
	
	protected Events()
	{
		plugin = RecipeManager.getPlugin();
		economy = RecipeManager.getEconomy();
		recipes = RecipeManager.getRecipes();
		settings = RecipeManager.getSettings();
		
		Bukkit.getPluginManager().registerEvents(this, plugin);
		
		for(final World world : Bukkit.getServer().getWorlds())
		{
			worldLoad(world);
		}
	}
	
	@EventHandler(priority = EventPriority.MONITOR)
	public void eventPlayerQuit(final PlayerQuitEvent event)
	{
		plugin.playerPage.remove(event.getPlayer().getName());
	}
	
	@EventHandler(priority = EventPriority.LOW, ignoreCancelled = true)
	public void eventInventoryClick(final InventoryClickEvent event)
	{
		try
		{
			final Inventory inv = event.getInventory();
			
			if(inv == null || !(inv instanceof FurnaceInventory))
				return;
			
			final InventoryHolder holder = inv.getHolder();
			
			if(holder == null || !(holder instanceof Furnace))
				return;
			
			final Furnace furnace = (Furnace)holder;
			final Player player = (Player)event.getWhoClicked();
			
			switch(event.getRawSlot())
			{
				case 0: // INGREDIENT slot
				{
					furnacePlaceIngredient(event, furnace, player, event.getCursor());
					return;
				}
				
				case 1: // FUEL slot
				{
					furnacePlaceFuel(event, furnace, player, event.getCursor());
					return;
				}
				
				case 2:
					return; // Result slot
					
				default: // player inventory
				{
					if(event.isShiftClick())
					{
						event.setCancelled(true);
						final ItemStack clicked = event.getCurrentItem();
						
						if(clicked == null) // clicked empty slot
							return;
						
						final int slot = event.isRightClick() ? 1 : 0;
						final ItemStack item = inv.getItem(slot);
						
						if(item == null) // no ingredient, place entire clicked stack
						{
							if(slot == 1 ? furnacePlaceFuel(event, furnace, player, clicked) : furnacePlaceIngredient(event, furnace, player, clicked))
							{
								inv.setItem(slot, clicked);
								event.setCurrentItem(null);
							}
							
							return;
						}
						
						final Material itemType = item.getType();
						final int itemStack = itemType.getMaxStackSize();
						final int itemAmount = item.getAmount();
						
						if(itemAmount < itemStack && itemType == clicked.getType() && item.getDurability() == clicked.getDurability()) // ingredient has room for more in the stack and it's the same type and data
						{
							final int amount = itemAmount + clicked.getAmount();
							final int diff = amount - itemStack;
							
							item.setAmount(Math.min(amount, itemStack));
							
							if(diff > 0)
								clicked.setAmount(diff);
							else
								event.setCurrentItem(null);
						}
						
						return;
					}
				}
			}
		}
		catch(final Exception e)
		{
			event.setCancelled(true);
			Messages.log(ChatColor.RED + event.getEventName() + " cancelled due to error:");
			e.printStackTrace();
			
			if(event.getWhoClicked() instanceof Player)
				((Player)event.getWhoClicked()).sendMessage(ChatColor.RED + event.getEventName() + " cancelled due to error!");
		}
	}
	
	private boolean furnacePlaceIngredient(final InventoryClickEvent event, final Furnace furnace, final Player player, final ItemStack item)
	{
		if(!plugin.allowedToCraft(player))
		{
			event.setCancelled(true);
			event.setResult(Result.DENY);
			return false;
		}
		
		final Smelt recipe = recipes.getSmeltRecipe(item);
		
		if(recipe == null) // no custom recipe, no reason to restrict
			return true;
		
		if(!recipe.usableBy(player, true))
		{
			event.setCancelled(true);
			event.setResult(Result.DENY);
			return false;
		}
		
		recipes.smelters.put(Recipes.locationToString(furnace.getLocation()), player.getName());
		
		return true;
	}
	
	private boolean furnacePlaceFuel(final InventoryClickEvent event, final Furnace furnace, final Player player, final ItemStack item)
	{
		if(!plugin.allowedToCraft(player))
		{
			event.setCancelled(true);
			event.setResult(Result.DENY);
			return false;
		}
		
		final Fuel recipe = recipes.getFuelRecipe(item);
		
		if(recipe == null) // no custom recipe, no reason to restrict
			return true;
		
		if(!recipe.usableBy(player, true))
		{
			event.setCancelled(true);
			event.setResult(Result.DENY);
			return false;
		}
		
		final String loc = Recipes.locationToString(furnace.getLocation());
		
		recipes.fuelers.put(loc, player.getName());
		recipes.fueled.put(loc, recipe.getFuel());
		
		return true;
	}
	
	@EventHandler(priority = EventPriority.LOW, ignoreCancelled = true)
	public void eventPreCraft(final PrepareItemCraftEvent event)
	{
		try
		{
			final Player player = (event.getView() == null ? null : (Player)event.getView().getPlayer());
			
			if(player != null && !plugin.allowedToCraft(player))
			{
				event.getInventory().setItem(0, null);
				return;
			}
			
			if(event.isRepair())
			{
				if(!settings.REPAIR_RECIPES)
					event.getInventory().setItem(0, null);
				
				return;
			}
			
			final Recipe craftedRecipe = event.getRecipe();
			
			if(craftedRecipe == null)
				return;
			
			final ItemStack craftedResult = craftedRecipe.getResult();
			
			if(!recipes.isCustomRecipe(craftedResult))
				return;
			
			if(craftedRecipe instanceof ShapedRecipe)
			{
				final Craft recipe = recipes.getCraftRecipes().get(craftedResult.getAmount());
				
				if(recipe == null)
					return;
				
				ItemStack result = null;
				
				if(player == null || recipe.usableBy(player, true))
					result = recipe.getResults().get(0).getItemStack(); // Display the first result item
					
				final RecipeManagerPreCraftEvent callEvent = new RecipeManagerPreCraftEvent(recipe, result, player);
				Bukkit.getPluginManager().callEvent(callEvent);
				
				result = callEvent.setDisplayResult();
				
				event.getInventory().setItem(0, result);
			}
			else if(craftedRecipe instanceof ShapelessRecipe)
			{
				final Combine recipe = recipes.getCombineRecipes().get(craftedResult.getAmount());
				
				if(recipe == null)
					return;
				
				ItemStack result = null;
				
				if(player == null || recipe.usableBy(player, true))
					result = recipe.getResults().get(0).getItemStack(); // Display the first result item
					
				final RecipeManagerPreCraftEvent callEvent = new RecipeManagerPreCraftEvent(recipe, result, player);
				Bukkit.getPluginManager().callEvent(callEvent);
				
				result = callEvent.setDisplayResult();
				
				event.getInventory().setItem(0, result);
			}
		}
		catch(final Exception e)
		{
			event.getInventory().setItem(0, null);
			Messages.log(ChatColor.RED + event.getEventName() + " cancelled due to error:");
			e.printStackTrace();
			
			if(event.getView() != null && event.getView().getPlayer() instanceof Player)
				((Player)event.getView().getPlayer()).sendMessage(ChatColor.RED + event.getEventName() + " cancelled due to error!");
		}
	}
	
	@SuppressWarnings("deprecation")
	@EventHandler(priority = EventPriority.LOW, ignoreCancelled = true)
	public void eventCraft(final CraftItemEvent event)
	{
		try
		{
			if(event.getCurrentItem() == null || event.getCurrentItem().getType() == Material.AIR)
				return;
			
			final Recipe r = event.getRecipe();
			
			if(r == null)
				return;
			
			final ItemStack recipeResult = r.getResult();
			
			if(!recipes.isCustomRecipe(recipeResult))
				return;
			
			final Player player = (event.getWhoClicked() == null ? null : (Player)event.getWhoClicked());
			
			if(r instanceof ShapedRecipe)
			{
				final Craft recipe = recipes.getCraftRecipes().get(recipeResult.getAmount());
				
				if(recipe == null)
					return;
				
				if(player != null && event.isShiftClick() && (recipe.getGiveExp() != null || recipe.getGiveLevel() != null || (economy.isEnabled() && recipe.getGiveMoney() != null)))
				{
					Messages.NOSHIFTCLICK_REWARDS.print(player);
					event.setCancelled(true);
					return;
				}
				
				Item result = null;
				
				if(recipe.getResults().size() > 1)
				{
					if(player != null && event.isShiftClick())
					{
						Messages.NOSHIFTCLICK_MULTIPLERESULTS.print(player);
						event.setCancelled(true);
						return;
					}
					
					result = recipe.getResult();
					
					final RecipeManagerCraftEvent callEvent = new RecipeManagerCraftEvent(recipe, result, player);
					Bukkit.getPluginManager().callEvent(callEvent);
					
					if(callEvent.isCancelled())
					{
						event.setCancelled(true);
						return;
					}
					
					result = callEvent.getResult();
					
					final ItemStack cursor = event.getCursor();
					
					if(player != null && (result.getType() == 0 || (cursor.getTypeId() != 0 && !result.compareItemStack(cursor))))
					{
						Messages.CRAFT_FAILURE.print(player, null, new String[][]
						{
							{
								"{chance}",
								result.getChance() + "%"
							}
						});
						
						final CraftingInventory inv = event.getInventory();
						ItemStack item;
						int amt;
						
						for(int i = 1; i < 10; i++)
						{
							item = inv.getItem(i);
							
							if(item != null)
							{
								if((amt = (item.getAmount() - 1)) > 0)
									item.setAmount(amt);
								else
									inv.clear(i);
							}
						}
						
						event.setResult(Result.DENY);
						event.setCancelled(true);
						
						if(result.getType() != 0)
						{
							final HashMap<Integer, ItemStack> extra = player.getInventory().addItem(cursor);
							player.updateInventory();
							
							if(extra != null && extra.size() > 0)
							{
								Messages.CRAFT_DROPPED.print(player);
								
								for(final ItemStack i : extra.values())
								{
									player.getWorld().dropItem(player.getLocation(), i);
								}
							}
							
							event.setCursor(result.getItemStack());
						}
					}
					else
						event.setCurrentItem(result.getItemStack());
				}
				else
				{
					result = recipe.getResults().get(0);
					
					final RecipeManagerCraftEvent callEvent = new RecipeManagerCraftEvent(recipe, result, player);
					Bukkit.getPluginManager().callEvent(callEvent);
					
					if(callEvent.isCancelled())
					{
						event.setCancelled(true);
						return;
					}
					
					result = callEvent.getResult();
					
					event.setCurrentItem(result.getItemStack());
				}
				
				/*
				 * Thanks to Comphenix for the research on this!
				 */
				/*
				 * TODO: finish this
				 * TODO: re-check exp/level/money each iteration of amount... might not have enought for all results...
				if(event.isShiftClick())
				{
					final Item res = result;
					final ItemStack[] items = player.getInventory().getContents();
					final ItemStack[] preItems = new ItemStack[items.length];
					
					for(int i = 0; i < items.length; i++)
					{
						if(items[i] != null)
							preItems[i] = items[i].clone();
					}
					
					Bukkit.getScheduler().scheduleSyncDelayedTask(plugin, new Runnable()
					{
						@Override
						public void run()
						{
							final ItemStack[] postItems = player.getInventory().getContents();
							int amount = 0;
							
							for(int i = 0; i < postItems.length; i++)
							{
								if(postItems[i] != null && postItems[i].getTypeId() == res.getType() && postItems[i].getDurability() == res.getData() && Objects.equal(postItems[i].getEnchantments(), res.getEnchantments()))
								{
									if(preItems[i] == null)
										amount += postItems[i].getAmount();
									else
										amount += (postItems[i].getAmount() - preItems[i].getAmount());
								}
							}
							
							for(int i = 0; i < amount; i++)
							{
								recipe.affectExp(player);
								recipe.affectLevel(player);
								recipe.affectMoney(player);
							}
							
							recipe.sendLog(player, "SHIFT+CLICK CRAFT x" + amount, null, res);
						}
					});
				}
				*/
				
				if(player != null)
				{
					recipe.affectExp(player);
					recipe.affectLevel(player);
					recipe.affectMoney(player);
					recipe.sendMessages(player, null, result);
				}
				
				recipe.sendCommands(player, null, result);
				recipe.sendLog((player == null ? null : player.getName()), result);
			}
			else if(r instanceof ShapelessRecipe)
			{
				final Combine recipe = recipes.getCombineRecipes().get(recipeResult.getAmount());
				
				if(recipe == null)
					return;
				
				if(event.isShiftClick() && (recipe.getGiveExp() != null || recipe.getGiveLevel() != null || (economy.isEnabled() && recipe.getGiveMoney() != null) || recipe.getCommands() != null))
				{
					Messages.NOSHIFTCLICK_REWARDS.print(player);
					event.setCancelled(true);
					return;
				}
				
				Item result = null;
				
				if(recipe.getResults().size() > 1)
				{
					if(player != null && event.isShiftClick())
					{
						Messages.NOSHIFTCLICK_MULTIPLERESULTS.print(player);
						event.setCancelled(true);
						return;
					}
					
					result = recipe.getResult();
					final ItemStack cursor = event.getCursor();
					
					if(player != null && (result.getType() == 0 || (cursor.getTypeId() != 0 && !result.compareItemStack(cursor))))
					{
						Messages.CRAFT_FAILURE.print(player, null, new String[][]
						{
							{
								"{chance}",
								result.getChance() + "%"
							}
						});
						
						final CraftingInventory inv = event.getInventory();
						ItemStack item;
						int amt;
						
						for(int i = 1; i < 10; i++)
						{
							item = inv.getItem(i);
							
							if(item != null)
							{
								if((amt = (item.getAmount() - 1)) > 0)
									item.setAmount(amt);
								else
									inv.clear(i);
							}
						}
						
						event.setResult(Result.DENY);
						event.setCancelled(true);
						
						if(result.getType() != 0)
						{
							final HashMap<Integer, ItemStack> extra = player.getInventory().addItem(cursor);
							player.updateInventory();
							
							if(extra != null && extra.size() > 0)
							{
								Messages.CRAFT_DROPPED.print(player);
								
								for(final ItemStack i : extra.values())
								{
									player.getWorld().dropItem(player.getLocation(), i);
								}
							}
							
							event.setCursor(result.getItemStack());
						}
					}
					else
						event.setCurrentItem(result.getItemStack());
				}
				else
					result = recipe.getResults().get(0);
				
				Bukkit.getPluginManager().callEvent(new RecipeManagerCraftEvent(recipe, result, player));
				
				if(player != null)
				{
					recipe.affectExp(player);
					recipe.affectLevel(player);
					recipe.affectMoney(player);
					recipe.sendMessages(player, null, result);
				}
				
				recipe.sendCommands(player, null, result);
				recipe.sendLog((player == null ? null : player.getName()), result);
			}
			else
				return;
			
			if(player != null)
			{
				// temp bugfix for Bukkit with shift clicking and returned items
				Bukkit.getScheduler().scheduleSyncDelayedTask(plugin, new Runnable()
				{
					@Override
					public void run()
					{
						player.updateInventory();
					}
				});
			}
		}
		catch(final Exception e)
		{
			event.setCurrentItem(null);
			event.setCancelled(true);
			Messages.log(ChatColor.RED + event.getEventName() + " cancelled due to error:");
			e.printStackTrace();
			
			if(event.getView() != null && event.getView().getPlayer() instanceof Player)
				((Player)event.getView().getPlayer()).sendMessage(ChatColor.RED + event.getEventName() + " cancelled due to error!");
		}
	}
	
	@EventHandler(priority = EventPriority.LOW, ignoreCancelled = true)
	public void eventFurnaceSmelt(final FurnaceSmeltEvent event)
	{
		try
		{
			final Smelt recipe = recipes.getSmeltRecipe(event.getSource());
			
			if(recipe == null)
				return;
			
			final Block block = event.getBlock();
			final Location location = block.getLocation();
			final String playerName = recipes.getFurnaceSmelter(location);
			final Player player = (playerName == null ? null : Bukkit.getPlayer(playerName));
			final String fuelerName = recipes.getFurnaceFueler(location);
			final Fuel fuelRecipe = recipes.getFuelRecipe(recipes.getFurnaceFuel(location));
			final Item result = (recipe.getChanceResult() ? recipe.getResult() : null);
			
			final RecipeManagerSmeltEvent callEvent = new RecipeManagerSmeltEvent(recipe, fuelRecipe, result, block, playerName, fuelerName);
			Bukkit.getPluginManager().callEvent(callEvent);
			
			if(callEvent.isCancelled())
			{
				event.setCancelled(true);
				return;
			}
			
			if(result == null || result.getType() == 0) // special handle for AIR results
			{
				final BlockState blockState = event.getBlock().getState();
				
				if(!(blockState instanceof Furnace))
					return;
				
				event.setCancelled(true); // setting the result to null will make the client timeout and the server spit errors
				final Furnace furnace = (Furnace)blockState;
				final ItemStack item = furnace.getInventory().getSmelting();
				final int amount = item.getAmount() - 1;
				
				if(amount <= 0)
					furnace.getInventory().clear(0);
				else
					item.setAmount(amount);
			}
			
			final ItemData ingredient = new ItemData(event.getSource());
			
			recipe.affectExp(player);
			recipe.affectLevel(player);
			recipe.affectMoney(player);
			recipe.sendCommands(player, ingredient, result);
			recipe.sendMessages(player, ingredient, result);
			recipe.sendLog(playerName, ingredient, result);
		}
		catch(final Exception e)
		{
			event.setCancelled(true);
			Messages.log(ChatColor.RED + event.getEventName() + " cancelled due to error:");
			e.printStackTrace();
		}
	}
	
	@EventHandler(ignoreCancelled = true)
	public void eventFurnaceBurn(final FurnaceBurnEvent event)
	{
		try
		{
			final Block block = event.getBlock();
			final Location loc = block.getLocation();
			
			if(!recipes.furnaces.containsKey(loc))
				recipes.furnaces.put(Recipes.locationToString(loc), (double)0);
			
			final ItemStack ingredient = event.getFuel();
			final Fuel recipe = recipes.getFuelRecipe(ingredient);
			
			if(recipe != null)
			{
				final String playerName = recipes.getFurnaceFueler(block.getLocation());
				final Player player = (playerName == null ? null : Bukkit.getPlayer(playerName));
				
				final RecipeManagerFuelBurnEvent callEvent = new RecipeManagerFuelBurnEvent(recipe, (20 * recipe.getTime()), block, playerName);
				Bukkit.getPluginManager().callEvent(callEvent);
				
				if(callEvent.isCancelled())
				{
					event.setCancelled(true);
					return;
				}
				
				event.setBurnTime(callEvent.getBurnTicks());
				event.setBurning(true);
				
				recipe.affectExp(player);
				recipe.affectLevel(player);
				recipe.affectMoney(player);
				recipe.sendCommands(player, recipe.getFuel(), null);
				recipe.sendMessages(player, recipe.getFuel(), null);
				recipe.sendLog(playerName, recipe.getFuel());
			}
			
			if(settings.FUEL_RETURN_BUCKETS)
			{
				switch(ingredient.getType())
				{
					case WATER_BUCKET:
					case LAVA_BUCKET:
					case MILK_BUCKET:
					{
						Bukkit.getScheduler().scheduleSyncDelayedTask(plugin, new Runnable()
						{
							@Override
							public void run()
							{
								final BlockState blockState = event.getBlock().getState();
								
								if(blockState instanceof Furnace)
									((Furnace)blockState).getInventory().setItem(1, new ItemStack(Material.BUCKET, 1));
							}
						});
						
						break;
					}
				}
			}
		}
		catch(final Exception e)
		{
			event.setCancelled(true);
			Messages.log(ChatColor.RED + event.getEventName() + " cancelled due to error:");
			e.printStackTrace();
		}
	}
	
	// Monitor furnaces
	
	@EventHandler(priority = EventPriority.MONITOR, ignoreCancelled = true)
	public void eventBlockBreak(final BlockBreakEvent event)
	{
		final Block block = event.getBlock();
		
		switch(block.getType())
		{
			case BURNING_FURNACE:
			case FURNACE:
			{
				final String locString = Recipes.locationToString(block.getLocation());
				
				recipes.furnaces.remove(locString);
				recipes.smelters.remove(locString);
				recipes.fuelers.remove(locString);
				recipes.fueled.remove(locString);
				
				break;
			}
		}
	}
	
	@EventHandler(priority = EventPriority.MONITOR, ignoreCancelled = true)
	public void eventWorldLoad(final WorldLoadEvent event)
	{
		worldLoad(event.getWorld());
	}
	
	@EventHandler(priority = EventPriority.MONITOR, ignoreCancelled = true)
	public void eventChunkLoad(final ChunkLoadEvent event)
	{
		if(!event.isNewChunk())
			furnaceChunk(event.getChunk(), true);
	}
	
	@EventHandler(priority = EventPriority.MONITOR, ignoreCancelled = true)
	public void eventChunkUnload(final ChunkUnloadEvent event)
	{
		furnaceChunk(event.getChunk(), false);
	}
	
	private void worldLoad(final World world)
	{
		final Chunk chunks[] = world.getLoadedChunks();
		
		for(final Chunk chunk : chunks)
		{
			furnaceChunk(chunk, true);
		}
	}
	
	private void furnaceChunk(final Chunk chunk, final boolean add)
	{
		if(chunk == null)
			return;
		
		final BlockState tileEnts[] = chunk.getTileEntities();
		
		if(tileEnts.length <= 0)
			return;
		
		for(final BlockState ent : tileEnts)
		{
			if(ent != null && ent instanceof Furnace)
			{
				if(add)
				{
					if(ent.getType() == Material.BURNING_FURNACE)
						recipes.furnaces.put(Recipes.locationToString(ent.getLocation()), (double)0);
				}
				else
					recipes.furnaces.remove(Recipes.locationToString(ent.getLocation()));
			}
		}
	}
}