package digi.recipeManager;

import java.io.*;
import java.util.*;
import java.util.Map.Entry;
import java.util.regex.*;

import org.bukkit.*;
import org.bukkit.command.CommandSender;
import org.bukkit.configuration.file.*;
import org.bukkit.plugin.java.JavaPlugin;

import digi.recipeManager.Metrics.Graph;
import digi.recipeManager.data.*;

/**
 * RecipeManager's main class which is also the static pointer for the API.<br>
 * <b>You should NOT create a new instance of this class! Access it staticly.</b>
 * 
 * @version 1.22
 * @author Digi
 * @see <a href="http://dev.bukkit.org/server-mods/recipemanager/">BukkitDev page</a>
 */
public final class RecipeManager extends JavaPlugin
{
	// TODO: Note to self: remember to always change these when updating the respective files
	private static final String				LAST_CHANGED_SETTINGS	= "1.22";
	private static final String				LAST_CHANGED_README		= "1.22";
	private static final String				LAST_CHANGED_ALIASES	= "1.21";
	protected static final String			LAST_CHANGED_MESSAGES	= "1.22";
	
	protected HashMap<String, Page>			playerPage				= new HashMap<String, Page>();
	
	private final HashMap<String, String>	itemAliases				= new HashMap<String, String>();
	private String							pluginVersion;
	
	protected static RecipeManager			plugin;
	protected static Settings				settings;
	protected static Recipes				recipes;
	protected static Econ					economy;
	
	/**
	 * <b>You should NOT trigger this manually!</b>
	 */
	@SuppressWarnings("unchecked")
	@Override
	public void onEnable()
	{
		plugin = this;
		settings = new Settings();
		recipes = new Recipes();
		economy = new Econ();
		
		// load .dat files
		
		Object loaded = loadObject("smelters.dat");
		
		if(loaded != null)
			recipes.smelters = (HashMap<String, String>)loaded;
		
		loaded = loadObject("fuelers.dat");
		
		if(loaded != null)
			recipes.fuelers = (HashMap<String, String>)loaded;
		
		loaded = loadObject("fueled.dat");
		
		if(loaded != null)
			recipes.fueled = (HashMap<String, ItemData>)loaded;
		
		// Register commands
		
		getCommand("rmrecipes").setExecutor(new CmdRecipes());
		getCommand("rmcheck").setExecutor(new CmdCheck());
		getCommand("rmreload").setExecutor(new CmdReload());
		
		// Execute task 0.5 seconds after all plugins have loaded
		
		getServer().getScheduler().scheduleSyncDelayedTask(this, new Runnable()
		{
			@Override
			public void run()
			{
				Bukkit.getLogger().fine("Loading settings and recipes...");
				
				loadSettings();
				recipes.loadRecipes(false);
				new Events();
				metrics();
				
				Bukkit.getLogger().fine("Done.");
			}
		}, 10);
	}
	
	/**
	 * <b>You should NOT trigger this manually!</b>
	 */
	@Override
	public void onDisable()
	{
		Bukkit.getScheduler().cancelTasks(plugin);
		recipes.furnaces.clear();
		recipes.removeCustomRecipes();
		recipes.reset();
		
		saveObject(recipes.smelters, "smelters.dat");
		saveObject(recipes.fuelers, "fuelers.dat");
		saveObject(recipes.fueled, "fueled.dat");
	}
	
	private void saveObject(final Object map, final String fileName)
	{
		try
		{
			final ObjectOutputStream stream = new ObjectOutputStream(new FileOutputStream(getDataFolder() + File.separator + fileName));
			stream.writeObject(map);
			stream.flush();
			stream.close();
		}
		catch(final Exception e)
		{
			Messages.log(ChatColor.RED + "Couldn't save " + fileName + " due to an error:");
			e.printStackTrace();
		}
	}
	
	private Object loadObject(final String fileName)
	{
		final File file = new File(getDataFolder() + File.separator + fileName);
		
		if(file.exists())
		{
			try
			{
				final ObjectInputStream stream = new ObjectInputStream(new FileInputStream(file));
				final Object result = stream.readObject();
				stream.close();
				return result;
			}
			catch(final Exception e)
			{
				Messages.log(ChatColor.RED + "Couldn't load " + fileName + " due to an error:");
				e.printStackTrace();
			}
		}
		
		return null;
	}
	
	/**
	 * Get the instance of the plugin to access it's methods.
	 * 
	 * @return The RecipeManager class
	 */
	public static RecipeManager getPlugin()
	{
		return plugin;
	}
	
	/**
	 * Returns the methods for recipes (not the recipes list).
	 * 
	 * @return the Recipes class
	 */
	public static Recipes getRecipes()
	{
		return recipes;
	}
	
	/**
	 * Returns the plugin's settings from config.yml
	 * 
	 * @return Config class for settings
	 */
	public static Settings getSettings()
	{
		return settings;
	}
	
	/**
	 * @return
	 */
	public static Econ getEconomy()
	{
		return economy;
	}
	
	/**
	 * Gets the version string of RecipeManager
	 * 
	 * @return
	 */
	public String getVersion()
	{
		return pluginVersion;
	}
	
	/**
	 * Gets a hashmap of item aliases defined in aliases.yml
	 * 
	 * @return
	 */
	public HashMap<String, String> getAliases()
	{
		return itemAliases;
	}
	
	/**
	 * Checks player for "recipemanager.craft" permission
	 * 
	 * @param sender
	 * @return
	 */
	public boolean allowedToCraft(final CommandSender sender)
	{
		if(!sender.hasPermission("recipemanager.craft"))
			return false;
		
		return true;
	}
	
	/* ======================================================================================== */
	
	protected void loadSettings()
	{
		if(!getDataFolder().exists())
			Bukkit.getLogger().warning("The 'RecipeManager' folder was NOT found, this means you didn't extract it from the .zip archive, which you should have done.");
		
		// Load: config.yml
		
		File file = new File(getDataFolder() + File.separator + "config.yml");
		
		if(!file.exists())
		{
			saveResource("config.yml", false);
			Messages.log("<green>config.yml file created, you should configure it!");
		}
		
		reloadConfig();
		settings.load(getConfig());
		
		String version = plugin.getConfig().getString("lastchanged", null);
		
		if(version == null || !version.equals(LAST_CHANGED_SETTINGS))
			Messages.log("<yellow>config.yml has changed! You should delete it, use rmreload to re-generate it and then re-configure it, and then rmreload again.");
		
		// Apply item return items
		
		net.minecraft.server.Item.WATER_BUCKET.a((settings.RETURN_BUCKETS ? net.minecraft.server.Item.BUCKET : null));
		net.minecraft.server.Item.LAVA_BUCKET.a((settings.RETURN_BUCKETS ? net.minecraft.server.Item.BUCKET : null));
		net.minecraft.server.Item.MILK_BUCKET.a((settings.RETURN_BUCKETS ? net.minecraft.server.Item.BUCKET : null));
		
		net.minecraft.server.Item.MUSHROOM_SOUP.a((settings.RETURN_BOWL ? net.minecraft.server.Item.BOWL : null));
		
		net.minecraft.server.Item.POTION.a((settings.RETURN_POTIONS ? net.minecraft.server.Item.GLASS_BOTTLE : null));
		net.minecraft.server.Item.EXP_BOTTLE.a((settings.RETURN_POTIONS ? net.minecraft.server.Item.GLASS_BOTTLE : null));
		
		// readme.txt
		
		file = new File(getDataFolder() + File.separator + "readme.txt");
		
		if(!file.exists())
		{
			saveResource("readme.txt", false);
			Messages.log("<green>readme.txt file created, you should read it to learn how to make recipes!");
		}
		else
		{
			try
			{
				final BufferedReader reader = new BufferedReader(new FileReader(file));
				final String line = reader.readLine();
				reader.close();
				
				final Matcher match = Pattern.compile("v(.*?) ").matcher(line);
				boolean replace = true;
				
				if(match.find())
				{
					version = line.substring(match.start() + 1, match.end()).trim();
					
					if(version.equals(LAST_CHANGED_README))
						replace = false;
				}
				
				if(replace)
				{
					saveResource("readme.txt", true);
					Messages.log("<yellow>readme.txt has changed, overwrited with new one, you should re-read it!");
				}
			}
			catch(final Exception e)
			{
				e.printStackTrace();
			}
		}
		
		// Load: aliases.yml
		
		file = new File(getDataFolder() + File.separator + "aliases.yml");
		
		if(!file.exists())
		{
			saveResource("aliases.yml", false);
			Messages.log("aliases.yml file created.");
		}
		
		final FileConfiguration aliases = YamlConfiguration.loadConfiguration(file);
		
		version = aliases.getString("lastchanged", null);
		
		if(version == null || !version.equals(LAST_CHANGED_ALIASES))
			Messages.log("<yellow>aliases.yml has changed! You should delete it, use rmreload to re-generate it and then re-configure it, and then rmreload again.");
		
		itemAliases.clear();
		String itemStr;
		
		for(final Entry<String, Object> entry : aliases.getValues(false).entrySet())
		{
			if(entry.getKey().equals("lastchanged"))
				continue;
			
			itemStr = entry.getValue().toString().toUpperCase();
			
			if(!itemStr.contains(":"))
				itemStr = itemStr + ":*";
			
			itemAliases.put(entry.getKey().toUpperCase(), itemStr);
		}
		
		// Load: messages.yml
		
		Messages.loadMessages();
	}
	
	private void metrics()
	{
		try
		{
			final Metrics metrics = new Metrics(plugin);
			Graph graph;
			
			// Graphs for total recipes
			
			graph = metrics.createGraph("Custom recipes");
			
			graph.addPlotter(new Metrics.Plotter("Craft recipes")
			{
				@Override
				public int getValue()
				{
					return recipes.getCraftRecipes().size();
				}
			});
			
			graph.addPlotter(new Metrics.Plotter("Combine recipes")
			{
				@Override
				public int getValue()
				{
					return recipes.getCombineRecipes().size();
				}
			});
			
			graph.addPlotter(new Metrics.Plotter("Smelt recipes")
			{
				@Override
				public int getValue()
				{
					return recipes.getSmeltRecipes().size();
				}
			});
			
			graph.addPlotter(new Metrics.Plotter("Fuels")
			{
				@Override
				public int getValue()
				{
					return recipes.getFuels().size();
				}
			});
			
			/*
			metrics.addCustomData(new Metrics.Plotter("Recipe syntax errors")
			{
				@Override
				public int getValue()
				{
					return recipes.getRecipeErrors().size();
				}
			});
			*/
			
			metrics.start();
		}
		catch(final IOException e)
		{
			Bukkit.getLogger().info("(Not really important) Metrics error: " + e.getMessage());
		}
	}
}